unit KDaoTable;
//************************************************************************************************
//                               Delphi Dao Project Version 1.0
//                             Copyright (c) 2000 by Kiril Antonov
//
//
//************************************************************************************************
{$DEFINE DAO35}
interface
uses
{$IFDEF DAO35}
DAO35Api,
{$ENDIF}
{$IFDEF DAO36}
DAO36Api,
{$ENDIF}
Windows,SysUtils,Classes,Db,DBTables,KDaoDataBase,DsgnIntf,ActiveX,Forms;

const
        MYBOOKMARKSIZE=8;


type
TDaoInfo=record
        RecordNo        : Integer;
        RecordData      : TStringList;
        BookmarkFlag    : TBookmarkFlag;
        BookmarkData    : Integer;
        DaoBookmark     : TSafeArray;
        BookmarkString  : array[0..MYBOOKMARKSIZE] of Char;
end;
PDaoInfo=^TDaoInfo;

TOO    = (
          dbDenyWrite,
          dbDenyRead,
          dbReadOnly,
          dbAppendOnly,
          dbInconsistent,
          dbConsistent,
          dbSQLPassThrough,
          dbFailOnError,
          dbForwardOnly,
          dbSeeChanges,
          dbRunAsync,
          dbExecDirect
          );
TOOSet = Set of TOO;


TTableNameEditor = class(TStringProperty)
    Public
      Procedure GetValues( Proc: TGetStrProc); override;
      Function  GetAttributes: TPropertyAttributes; override;
    End;

TQueryNameEditor = class(TStringProperty)
    Public
      Procedure GetValues( Proc: TGetStrProc); override;
      Function  GetAttributes: TPropertyAttributes; override;
    End;

TIndexNameEditor = class(TStringProperty)
    Public
      Procedure GetValues( Proc: TGetStrProc); override;
      Function  GetAttributes: TPropertyAttributes; override;
    End;

TTableTypeEditor = class(TIntegerProperty)
    Public
     Function  GetValue: string; override;
     Procedure GetValues( Proc: TGetStrProc);override;
     Procedure SetValue(const Value: string); override;
     Function  GetAttributes: TPropertyAttributes; override;
    End;

TLockTypeEditor = class(TIntegerProperty)
    Public
     Function  GetValue: string; override;
     Procedure GetValues( Proc: TGetStrProc);override;
     Procedure SetValue(const Value: string); override;
     Function  GetAttributes: TPropertyAttributes; override;
    End;



type TKADaoTable = class(TDataSet)
private
        F_Active      : Boolean;
        F_RecNo       : Integer;
        F_FilterBuffer: PChar;
        F_BufferSize  : Integer;
        F_StartMyInfo : Integer;
        F_StartCalc   : Integer;
        Function        GetActiveRecordBuffer:  PChar;
        Function        FilterRecord(Buffer: PChar): Boolean;
protected
        F_Database               : TKADaoDatabase;
        F_ReadOnly               : Boolean;
        F_Filtered               : Boolean;
        F_DaoTable               : Recordset;
        F_SQL                    : TStrings;
        F_Sort                   : TStrings;
        F_TableName              : String;
        F_QueryDefName           : String;
        F_QueryDefParameters     : TStrings;
        F_QueryDefSQLText        : TStrings;
        F_IndexName              : String;
        F_TableType              : Integer;
        F_LockType               : Integer;
        F_OpenOptions            : TOOSet;
        F_RecordSize             : Integer;
        Function        F_Get_Database:TKADaoDatabase;
        Procedure       F_Set_Database(Value:TKADaoDatabase);
        Function        F_Get_TableName:String;
        Procedure       F_Set_TableName(Value:String);
        Function        F_Get_IndexName:String;
        Procedure       F_Set_IndexName(Value:String);
        Procedure       F_Set_TableType(Value:Integer);
        Procedure       F_Set_LockType(Value:Integer);
        Procedure       F_Set_OpenOptions(Value:TOOSet);
        Procedure       F_Set_ReadOnly(Value:Boolean);
        Procedure       F_Set_Filtered(Value:Boolean);
        Procedure       F_SetSort(Value:TStrings);
        Procedure       F_SetSQL(Value:TStrings);
        Procedure       F_SetQueryDefName(Value:String);
        Procedure       F_SetQueryDefParameters(Value:TStrings);
        Procedure       F_SetQueryDefSQLText(Value:TStrings);


        Function        InternalCalcRecordSize:Integer;
        Function        IntegerToBuffer(Buffer: Pointer; S: String): Boolean;
        Function        FloatToBuffer(Buffer: Pointer; S: String): Boolean;
        Function        DateToBuffer(Buffer: Pointer; S: String): Boolean;
        Function        TimeToBuffer(Buffer: Pointer; S: String): Boolean;
        Function        DateTimeToBuffer(Buffer: Pointer; S: String): Boolean;
        Function        BooleanToBuffer(Buffer: Pointer; S: String): Boolean;

        Function        BufferToDate(Buffer: Pointer): String;
        Function        BufferToTime(Buffer: Pointer): String;
        Function        BufferToDateTime(Buffer: Pointer): String;


        Procedure       InternalOpen; override;
        Procedure       InternalClose; override;
        Function        IsCursorOpen: Boolean; override;
        Function        GetCanModify: Boolean; override;
        Function        GetRecordSize: Word;override;
        Function        AllocRecordBuffer: PChar; override;
        Procedure       FreeRecordBuffer(var Buffer: PChar); override;
        Procedure       InternalFirst;override;
        Procedure       InternalLast;override;
        Procedure       InternalInitFieldDefs; override;
        Procedure       InternalInitRecord(Buffer: PChar); override;
        Function        GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
        Procedure       SetFieldData(Field: TField; Buffer: Pointer);override;
        Procedure       ClearDBCalcFields(Buffer: PChar);

        Procedure       InternalCancel; override;
        Procedure       InternalEdit; override;
        Procedure       InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
        Procedure       InternalPost; override;
        Procedure       InternalDelete; override;
        Function        GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
        Procedure       InternalSetToRecord(Buffer: PChar); override;
        Procedure       InternalRefresh; override;

        Procedure       InternalGotoBookmark(Bookmark: Pointer); override;
        Function        GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
        Procedure       SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
        Function        GetBookmarkStr: TBookmarkStr; override;
        Procedure       SetBookmarkStr(const Value: TBookmarkStr); override;
        Procedure       GetBookmarkData(Buffer: PChar; Data: Pointer); override;
        Procedure       SetBookmarkData(Buffer: PChar; Data: Pointer); override;

        Procedure       InternalHandleException; override;
        Procedure       DataEvent(Event: TDataEvent; Info: Longint); override;

        Function        CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
        Function        GetRecordCount  : Integer;override;

  public
        MainDatabaseShutdown   : Boolean;
        Constructor              Create(AOwner: TComponent); override;
        Destructor               Destroy; override;
  published
        Property Database            : TKADaoDatabase Read F_Get_Database Write F_Set_Database;
        Property SortedBy            : TStrings Read F_Sort Write F_SetSort;
        Property TableName           : String Read F_Get_TableName Write F_Set_TableName;
        Property QueryDefName        : String Read F_QueryDefName Write F_SetQueryDefName;
        Property QueryDefParameters  : TStrings Read F_QueryDefParameters Write F_SetQueryDefParameters;
        Property QueryDefSQLText     : TStrings Read F_QueryDefSQLText Write F_SetQueryDefSQLText;
        Property SQL                 : TStrings Read F_SQL Write F_SetSQL;
        Property TableType           : Integer Read F_TableType Write F_Set_TableType;
        Property LockType            : Integer Read F_LockType Write F_Set_LockType;
        Property OpenOptions         : TOOSet Read F_OpenOptions Write F_Set_OpenOptions;
        Property IndexName           : String Read F_Get_IndexName Write F_Set_IndexName;
        Property ReadOnly            : Boolean Read F_ReadOnly Write F_Set_ReadOnly;
        Property Filtered            : Boolean Read F_Filtered Write F_Set_Filtered;
        Property Filter;
        Property BeforeOpen;
        Property AfterOpen;
        Property BeforeClose;
        Property AfterClose;
        Property BeforeInsert;
        Property AfterInsert;
        Property BeforeEdit;
        Property AfterEdit;
        Property BeforePost;
        Property AfterPost;
        Property BeforeCancel;
        Property AfterCancel;
        Property BeforeDelete;
        Property AfterDelete;
        Property BeforeScroll;
        Property AfterScroll;
        Property OnCalcFields;
        Property OnDeleteError;
        Property OnEditError;
        Property OnFilterRecord;
        Property OnNewRecord;
        Property OnPostError;
        Property Active;
end;

// Handle Memo fields
  TKBlobStream = class(TStream)
  private
    F_Field      : TBlobField;
    F_DataSet    : TKADaoTable;
    F_Buffer     : PChar;
    F_Mode       : TBlobStreamMode;
    F_Opened     : Boolean;
    F_Modified   : Boolean;
    F_Position   : Longint;
    F_BlobData   : TBlobData;
    F_BlobSize   : Integer;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    procedure Truncate;
  end;


Procedure Register;

implementation
Uses DaoUtils,Dialogs;

Const
  CRLF=#13+#10;

constructor TKADaoTable.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  F_TableName            := '';
  F_TableType            := dbOpenDynaset;
  F_LockType             := dbPessimistic;
  F_OpenOptions          := [dbInconsistent];
  F_ReadOnly             := False;
  F_Filtered             := False;
  MainDatabaseShutdown   := False;
  F_QueryDefName         := '';
  F_QueryDefSQLText      := TStringList.Create;
  F_QueryDefSQLText.Clear;
  F_SQL                  := TStringList.Create;
  F_SQL.Clear;
  F_Sort                 := TStringList.Create;
  F_Sort.Clear;
  F_QueryDefParameters   := TStringList.Create;
  F_QueryDefParameters.Clear;
  F_Database             := Nil;    
end;

destructor TKADaoTable.Destroy;
begin
  if F_Active Then
      Begin
        InternalClose;
        F_Active:=False;
      End;
  F_SQL.Free;
  F_Sort.Free;
  F_QueryDefParameters.Free;
  F_QueryDefSQLText.Free;
  inherited Destroy;
end;

Function TKADaoTable.F_Get_Database:TKADaoDatabase;
Begin
 if NOT Assigned(F_Database) Then F_Database:=Nil;
 Result:=F_Database;
End;

Procedure TKADaoTable.F_Set_Database(Value:TKADaoDatabase);
Begin
  if Active Then DatabaseError('TKADaoTable.F_Set_Database: Cannot set Database while table is Active!');
  F_Database:=Value;
End;

Function TKADaoTable.F_Get_TableName:String;
Begin
 Result:= F_TableName;
End;

Procedure TKADaoTable.F_Set_TableName(Value:String);
Begin
  if Active Then DatabaseError('Cannot set TableName while Table is active!');
  F_IndexName:='';
  F_TableName:=Value;
  F_SQL.Clear;
  F_QueryDefName:='';
  F_QueryDefParameters.Clear;
End;

Procedure TKADaoTable.F_SetSQL(Value:TStrings);
Begin
 F_SQL.SetText(Value.GetText);
 F_QueryDefParameters.Clear;
 F_QueryDefName:='';
 F_TableName:='';
End;

Procedure TKADaoTable.F_SetQueryDefName(Value:String);
Begin
  F_TableName:='';
  F_SQL.Clear;
  F_QueryDefParameters.Clear;
  F_QueryDefName:=Value;
  Try
    if Assigned(F_Database) And (F_Database.Connected) Then
       Begin
         F_QueryDefSQLText.Clear;
         F_QueryDefSQLText.SetText(PChar(F_Database.GetQueryDefSQLText(Value)));
       End;  
  Except
  End;
End;


Function TKADaoTable.F_Get_IndexName:String;
Begin
 Result:= F_IndexName;
End;

Procedure TKADaoTable.F_Set_IndexName(Value:String);
Begin
  if Active Then DatabaseError('Cannot set IndexName while Table is active!');
  F_IndexName:=Value;
End;

Procedure TKADaoTable.F_Set_TableType(Value:Integer);
Begin
  if Active Then DatabaseError('Cannot set TableType while Table is active!');
  F_TableType:=Value;
End;

Procedure TKADaoTable.F_Set_LockType(Value:Integer);
Begin
  if Active Then DatabaseError('Cannot set LockType while Table is active!');
  F_LockType:=Value;
End;

Procedure TKADaoTable.F_Set_OpenOptions(Value:TOOSet);
Begin
  if Active Then DatabaseError('Cannot set OpenOptions while Table is active!');
  F_OpenOptions:=Value;
End;

Procedure TKADaoTable.F_Set_ReadOnly(Value:Boolean);
Begin
  if Active Then DatabaseError('TKADaoTable.F_Set_ReadOnly: Cannot set ReadOnly while table is Active!');
  F_ReadOnly:=Value;
End;

Procedure TKADaoTable.F_Set_Filtered(Value:Boolean);
Begin
 If (F_Active) And (Filter <> '') And (Value) Then
    Begin
      F_DaoTable.Filter:=Filter;
    End
 Else
 If (F_Active) Then
    Begin
       F_DaoTable.Filter:='';
    End;
 F_Filtered:=Value;
End;

Procedure TKADaoTable.F_SetSort(Value:TStrings);
Begin
 F_Sort.SetText(Value.GetText);
End;

Procedure TKADaoTable.F_SetQueryDefParameters(Value:TStrings);
Begin
 F_QueryDefParameters.SetText(Value.GetText);
End;

Procedure TKADaoTable.F_SetQueryDefSQLText(Value:TStrings);
Begin
//*************************** READ ONLY
End;

Function TKADaoTable.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
Begin
  Result:=TKBlobStream.Create(Field As TBlobField,Mode);
End;

Function TKADaoTable.InternalCalcRecordSize:Integer;
Begin
 F_RecordSize:=0;
 Result:=F_RecordSize;
End;

Procedure TKADaoTable.InternalOpen;
Var
 X       : Integer;
 TabType : Integer;
 LoType  : Integer;
 Options : Integer;
 TempRS  : Recordset;
 TabN    : String;
begin
        if Not Assigned(F_Database) Then DatabaseError('TKADaoTable.InternalOpen: Missing Database');
        if (TableName='') And
           (SQL.Count=0)  And
           (QueryDefName='')
        Then DatabaseError('TKADaoTable.InternalOpen: Missing TableName or QueryDefName or SQL is empty!');

        if F_Database.Connected= False Then DatabaseError('TKADaoTable.InternalOpen: Database is NOT connected');
        If Name='' Then Name:='KADaoTable'+IntToStr(GetTickCount);

        TabType:=F_TableType;
        LoType:=F_LockType;
        Options:=0;
        {$IFDEF DAO35}
        if dbDenyWrite      in F_OpenOptions Then Options:=Options And DAO35Api.dbDenyWrite;
        if dbDenyRead       in F_OpenOptions Then Options:=Options And DAO35Api.dbDenyRead;
        if dbReadOnly       in F_OpenOptions Then Options:=Options And DAO35Api.dbReadOnly;
        if dbAppendOnly     in F_OpenOptions Then Options:=Options And DAO35Api.dbAppendOnly;
        if dbInconsistent   in F_OpenOptions Then Options:=Options And DAO35Api.dbInconsistent;
        if dbConsistent     in F_OpenOptions Then Options:=Options And DAO35Api.dbConsistent;
        if dbSQLPassThrough in F_OpenOptions Then Options:=Options And DAO35Api.dbSQLPassThrough;
        if dbFailOnError    in F_OpenOptions Then Options:=Options And DAO35Api.dbFailOnError;
        if dbForwardOnly    in F_OpenOptions Then Options:=Options And DAO35Api.dbForwardOnly;
        if dbSeeChanges     in F_OpenOptions Then Options:=Options And DAO35Api.dbSeeChanges;
        if dbRunAsync       in F_OpenOptions Then Options:=Options And DAO35Api.dbRunAsync;
        if dbExecDirect     in F_OpenOptions Then Options:=Options And DAO35Api.dbExecDirect;
        {$ENDIF}
        {$IFDEF DAO36}
        if dbDenyWrite      in F_OpenOptions Then Options:=Options And DAO36Api.dbDenyWrite;
        if dbDenyRead       in F_OpenOptions Then Options:=Options And DAO36Api.dbDenyRead;
        if dbReadOnly       in F_OpenOptions Then Options:=Options And DAO36Api.dbReadOnly;
        if dbAppendOnly     in F_OpenOptions Then Options:=Options And DAO36Api.dbAppendOnly;
        if dbInconsistent   in F_OpenOptions Then Options:=Options And DAO36Api.dbInconsistent;
        if dbConsistent     in F_OpenOptions Then Options:=Options And DAO36Api.dbConsistent;
        if dbSQLPassThrough in F_OpenOptions Then Options:=Options And DAO36Api.dbSQLPassThrough;
        if dbFailOnError    in F_OpenOptions Then Options:=Options And DAO36Api.dbFailOnError;
        if dbForwardOnly    in F_OpenOptions Then Options:=Options And DAO36Api.dbForwardOnly;
        if dbSeeChanges     in F_OpenOptions Then Options:=Options And DAO36Api.dbSeeChanges;
        if dbRunAsync       in F_OpenOptions Then Options:=Options And DAO36Api.dbRunAsync;
        if dbExecDirect     in F_OpenOptions Then Options:=Options And DAO36Api.dbExecDirect;
        {$ENDIF}

        TabN:=TableName;
        if F_SQL.Count > 0 Then
           Begin
             TabN:=F_SQL.GetText;
           End;
        if F_QueryDefName <> '' Then
           Begin
             TabN:=F_QueryDefName;
             if Database.CoreDatabase.QueryDefs.Item[TabN].Parameters.Count <> F_QueryDefParameters.Count Then
                DatabaseError('TKADaoTable.InternalOpen: Different number of QueryDef Parameters!'+CRLF+
                'Expected: '+IntToStr(Database.CoreDatabase.QueryDefs.Item[TabN].Parameters.Count));
             For X:=0 To F_QueryDefParameters.Count-1 do
                 Begin
                   Database.CoreDatabase.QueryDefs.Item[TabN].Parameters.Item[X].Value:=F_QueryDefParameters.Strings[X];
                 End;
           End;
        if (F_QueryDefName <> '') And (Database.CoreDatabase.QueryDefs.Item[TabN].Parameters.Count > 0) Then
            Begin
              Try
                F_QueryDefSQLText.Clear;
                F_QueryDefSQLText.SetText(PChar(F_Database.GetQueryDefSQLText(TabN)));
              Except
              End;
              F_DaoTable:=Database.CoreDatabase.QueryDefs.Item[TabN].OpenRecordset(TabType,Options,LoType);
            End
        Else
            Begin
              Try
                F_QueryDefSQLText.Clear;
                F_QueryDefSQLText.SetText(PChar(F_Database.GetQueryDefSQLText(TabN)));
              Except
              End;
              F_DaoTable:=Database.CoreDatabase.OpenRecordset(TabN,TabType,Options,LoType);
            End;
        if F_IndexName <> '' Then F_DaoTable.Index:=F_IndexName
          Else
        if F_Sort.Count > 0 Then
           Begin
            F_DaoTable.Sort:=F_Sort.CommaText;
            TempRS:=F_DaoTable;
            F_DaoTable:=TempRS.OpenRecordset(TabType,Options);
            TempRS.Close;
           End;
        if Self.Name='' Then Self.Name:='KAdoTable'+IntToStr(GetTickCount);   
        F_Database.ActiveTableNames.AddObject(Self.Name,Self);
        InternalInitFieldDefs;
        if DefaultFields then CreateFields;
        BindFields(True);
        F_RecNo:=-1;
        BookmarkSize:=MYBOOKMARKSIZE;
        F_StartMyInfo:=InternalCalcRecordSize;
        F_StartCalc:=F_StartMyInfo+SizeOf(TDaoInfo);
        F_BufferSize:=F_RecordSize+Sizeof(TDaoInfo)+CalcFieldsSize;
        //****************************************************************
        F_Active:=True;
        InternalFirst;
        //****************************************************************
end;

Procedure TKADaoTable.InternalClose;
begin

        if Not F_Active Then Exit;
        F_Active:=False;
        BindFields(False);
        if DefaultFields then DestroyFields;
        F_DaoTable.Close;
        if NOT MainDatabaseShutdown Then
           F_Database.ActiveTableNames.Delete(F_Database.ActiveTableNames.IndexOf(Self.Name))
        Else
           MainDatabaseShutdown:=False;
end;

Procedure TKADaoTable.InternalFirst;
begin
    Try
      F_DaoTable.MoveFirst;
      F_DaoTable.MovePrevious;
    Except
    End;
    F_RecNo:=-1;
end;

Procedure TKADaoTable.InternalLast;
begin
     Try
        OleVariant(F_DaoTable).MoveLast;
        F_DaoTable.MoveNext;
        F_RecNo:=F_DaoTable.RecordCount;
     Except
        F_RecNo:=-1;
     End;
end;

Procedure TKADaoTable.InternalSetToRecord(Buffer: PChar);
begin
        F_RecNo:=PDaoInfo(Buffer+F_StartMyInfo)^.RecordNo;
        Try
         F_DaoTable.MoveFirst;
         F_DaoTable.Move(F_RecNo,NULL);
         {
         ShowMessage(IntToStr(Pinteger(F_DaoTable.Bookmark^.PvData)^));
         IF (F_DaoTable.Bookmarkable) And (PDaoInfo(Buffer+F_StartMyInfo)^.DaoBookmark.pvData <> Nil) Then
             Begin
               F_DaoTable.Bookmark^:=PDaoInfo(Buffer+F_StartMyInfo)^.DaoBookmark;
             End;
         if Pinteger(F_DaoTable.Bookmark^.PvData)^ <> Pinteger(PDaoInfo(Buffer+F_StartMyInfo)^.DaoBookmark.pvData)^ Then
            Begin
              F_DaoTable.MoveFirst;
              F_DaoTable.Move(F_RecNo,NULL);
            End;
         }   
        Except
        End;
end;

Procedure TKADaoTable.InternalRefresh;
begin
     Resync([rmExact, rmCenter]);
end;


//*********************************************************************************** BOOKMARK FUNCTIONS
procedure TKADaoTable.InternalGotoBookmark(Bookmark: Pointer);
begin
     if (F_Active) And F_DaoTable.Bookmarkable Then F_DaoTable.Bookmark.pvData := PInteger(Bookmark);
end;


Function TKADaoTable.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
        Result:=PDaoInfo(Buffer+F_StartMyInfo).BookmarkFlag;
end;

procedure TKADaoTable.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
     PDaoInfo(Buffer + F_StartMyInfo)^.BookmarkFlag := Value;
end;


Function TKADaoTable.GetBookmarkStr: TBookmarkStr;
Begin
  Try
    if F_DaoTable.Bookmarkable Then
       Begin
         Result := TBookmarkStr(IntToStr(PInteger(F_DaoTable.Bookmark.pvData)^));
       End
    Else
       Result := '';
  Except
    Result := '';
  End;
End;

Procedure TKADaoTable.SetBookmarkStr(const Value: TBookmarkStr);
var
   iBook : Integer;
Begin
   if (Value = '') And (F_DaoTable.Bookmarkable) Then iBook := PInteger(F_DaoTable.Bookmark.pvData)^ else iBook := StrToInt(Value);
   if iBook <> PInteger(F_DaoTable.Bookmark.pvData)^ then
      begin
        PInteger(F_DaoTable.Bookmark.pvData)^ := iBook;
        InternalRefresh;
      end;
End;

procedure TKADaoTable.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
     PInteger(Data)^ := PDaoInfo(Buffer + F_StartMyInfo)^.BookmarkData;
end;

procedure TKADaoTable.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
     if Buffer <> nil then PDaoInfo(Buffer + F_StartMyInfo)^.BookmarkData := PInteger(Data)^;
end;

//*********************************************************************************** BOOKMARK FUNCTIONS


Procedure TKADaoTable.InternalInitFieldDefs;
Var
  X, Sz :Integer;
begin
        FieldDefs.Clear;
        with FieldDefs do
        begin
          For X:=0 To F_DaoTable.Fields.Count-1 do
              Begin
                Sz:=DaoSizeToBDESize(F_DaoTable.Fields.Item[X].Type_,F_DaoTable.Fields.Item[X].Size);
                Add(F_DaoTable.Fields.Item[X].Name,DaoToBDE(F_DaoTable.Fields.Item[X].Type_),Sz,False);
              End;
        end;
end;

Function TKADaoTable.GetActiveRecordBuffer:  PChar;
begin
        case State of
             dsBrowse      : if IsEmpty then Result := nil else Result := ActiveBuffer;
             dsCalcFields  : Result := CalcBuffer;
             dsFilter      : Result := F_FilterBuffer;
             dsEdit,
             dsInsert      : Result := ActiveBuffer;
        else Result:=nil;
        end;
end;

procedure TKADaoTable.InternalInitRecord(Buffer: PChar);
Var
  X:Integer;
begin
     FillChar(Buffer^, F_BufferSize, 0);
     PDaoInfo(Buffer+F_StartMyInfo)^.RecordData:=TStringList.Create;

     PDaoInfo(Buffer+F_StartMyInfo)^.DaoBookmark.pvData:=Nil;
     For X:=0 To F_DaoTable.Fields.Count-1 do
          Begin
            PDaoInfo(Buffer+F_StartMyInfo)^.RecordData.Add('');
          End;
     //************************************************************************* TEST METHOD
     InternalLast;
     PDaoInfo(Buffer+F_StartMyInfo)^.RecordNo:=F_RecNo;
     //************************************************************************* TEST METHOD

     //************************************************************************* OK METHOD
     //PDaoInfo(Buffer+F_StartMyInfo)^.RecordNo:=GetRecordCount;
     //************************************************************************* OK METHOD
end;




Procedure TKADaoTable.InternalHandleException;
begin
     Application.HandleException(Self);
end;

procedure TKADaoTable.DataEvent(Event: TDataEvent; Info: Longint);
begin
     inherited DataEvent(Event, Info);
     if (Event in [deDataSetScroll,
                   deUpdateState,
                   deDataSetChange]) and Active then
         Begin
                   InternalSetToRecord(ActiveBuffer);
         End;
end;


Procedure TKADaoTable.ClearDBCalcFields(Buffer: PChar);
begin
        FillChar(Buffer[F_StartCalc],CalcFieldsSize,0);
end;

Function TKADaoTable.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  SourceBuffer : PChar;
  Value        : TStringList;
  FieldNumber  : Integer;
  TempString   : String;
begin
        Result:=False;
        SourceBuffer:=GetActiveRecordBuffer;
        if (not F_Active) or (SourceBuffer=nil) then
           Begin
              Exit;
           End;
        if (Field.FieldKind=fkCalculated) or (Field.FieldKind=fkLookup) then
          begin
                Inc(SourceBuffer,F_StartCalc+Field.Offset);
                if {(SourceBuffer[0]=#0) or} (Buffer=nil) then
                  Begin
                    Exit;
                  End
                Else
                  Begin
                    Move(SourceBuffer[1], Buffer^, Field.DataSize);
                    Result:=True;
                  End;
          end
        else
          Begin
           Value :=PDaoInfo(PChar(SourceBuffer)+F_StartMyInfo)^.RecordData;
           FieldNumber:=Field.FieldNo-1;
           if (Value=Nil) Then Exit;
           if (Buffer = Nil)  Then
            Begin
              Result:=(Value.Strings[FieldNumber] <> '');
            End
               Else
            Begin
              Case Field.DataType of
                ftBytes     : Result:=IntegerToBuffer(Buffer,Value.Strings[FieldNumber]);
                ftInteger   : Result:=IntegerToBuffer(Buffer,Value.Strings[FieldNumber]);
                ftSmallint  : Result:=IntegerToBuffer(Buffer,Value.Strings[FieldNumber]);
                ftCurrency  : Result:=FloatToBuffer(Buffer,Value.Strings[FieldNumber]);
                ftFloat     : Result:=FloatToBuffer(Buffer,Value.Strings[FieldNumber]);
                ftDate      : Result:=DateToBuffer(Buffer,Value.Strings[FieldNumber]);
                ftString    : Begin
                                TempString:=Value.Strings[FieldNumber];
                                TempString:=TempString+#0;
                                CopyMemory(PChar(Buffer),PChar(TempString),Length(TempString));
                                Result := True;
                              End;
                ftAutoInc   : Result:=IntegerToBuffer(Buffer,Value.Strings[FieldNumber]);
                ftTime      : Result:=TimeToBuffer(Buffer,Value.Strings[FieldNumber]);
                ftDateTime  : Result:=DateTimeToBuffer(Buffer,Value.Strings[FieldNumber]);
                ftBoolean   : Result:=BooleanToBuffer(Buffer,Value.Strings[FieldNumber]);
              End;
            End;
          End;
end;

Procedure TKADaoTable.SetFieldData(Field: TField; Buffer: Pointer);
var
        DestinationBuffer: PChar;
        Tmp              : String;
        BTmp             : WordBool;
        BBTmp            : Boolean;
begin
        DestinationBuffer:=GetActiveRecordBuffer;
        if (Field.FieldKind=fkCalculated) or (Field.FieldKind=fkLookup) then //this is a calculated field
        begin
                Inc(DestinationBuffer,F_StartCalc+Field.Offset);
                Boolean(DestinationBuffer[0]):=(Buffer<>nil);
                if Boolean(DestinationBuffer[0]) then CopyMemory(@DestinationBuffer[1],Buffer,Field.DataSize);
        end
        Else
        Begin
          Tmp:='';
          Case Field.DataType of
            ftString     : Tmp := PChar(Buffer);
            ftSmallint   : Tmp := IntToStr(Integer(Buffer^));
            ftWord       : Tmp := IntToStr(Integer(Buffer^));
            ftAutoInc    : Tmp := IntToStr(Integer(Buffer^));
            ftInteger    : Tmp := IntToStr(Integer(Buffer^));

            ftBCD        : Tmp := FloatToStr(Double(Buffer^));
            ftCurrency   : Tmp := FloatToStr(Double(Buffer^));
            ftFloat      : Tmp := FloatToStr(Double(Buffer^));

            ftDate       : Tmp := BufferToDate(Buffer);
            ftTime       : Tmp := BufferToTime(Buffer);
            ftDateTime   : Tmp := BufferToDateTime(Buffer);
            ftBoolean    : Begin
                            BTmp := WordBool(Buffer^);
                            BBTmp:=Boolean(BTmp);
                             Case BBTmp Of
                                  True   : Tmp:='True';
                                  False  : Tmp:='False';
                             End;
                           End;
          End;
           PDaoInfo(DestinationBuffer+F_StartMyInfo)^.RecordData[Field.FieldNo-1]:=Tmp;
           DataEvent(deFieldChange, Longint(Field));
        End;
end;

procedure TKADaoTable.InternalCancel;
var
   bLast : Boolean;
begin
     if F_Active then
     begin
          bLast := (State = dsInsert);
          if bLast then Last;
     end;
end;

procedure TKADaoTable.InternalEdit;
begin
     inherited InternalEdit;
end;

procedure TKADaoTable.InternalAddRecord(Buffer: Pointer; Append: Boolean);
Var
 X      : Integer;
 RData  : TStringList;
 S      : String;
 P      : Integer;
begin
    InternalLast;
    RData:=PDaoInfo(PChar(Buffer)+F_StartMyInfo)^.RecordData;
    F_DaoTable.AddNew;
    For X:=0 to RData.Count-1 do
        Begin
          S:=RData.Strings[X];
          if (Fields[X].DataType=ftDate) or
             (Fields[X].DataType=ftTime) or
             (Fields[X].DataType=ftDateTime) Then
          Begin
           P:=Pos(' ',S);
           if P > 0 Then S:=Copy(S,1,P-1);
          End;
          if (F_DaoTable.Fields.Item[X].Attributes And dbUpdatableField > 0) Then
              Begin
                if (F_DaoTable.Fields.Item[X].Attributes And dbAutoIncrField) = 0 Then
                    Begin
                      if (F_DaoTable.Fields.Item[X].Attributes And dbSystemField) = 0 Then
                          Begin
                            if (Fields[X].IsBlob) And (S = '(Blob)')  Then
                            Else F_DaoTable.Fields.Item[X].Value:=S;
                          End;
                    End;
              End;
        End;
    OleVariant(F_DaoTable).Update;
end;

procedure TKADaoTable.InternalPost;
Var
 Buffer : PChar;
 X      : Integer;
 RData  : TStringList;
 S      : String;
 P      : Integer;
Begin
 CheckActive;
 if State = dsEdit then
  Begin
    Buffer:=GetActiveRecordBuffer;
    RData:=PDaoInfo(Buffer+F_StartMyInfo)^.RecordData;
    F_DaoTable.Edit;
    For X:=0 to RData.Count-1 do
        Begin
          S:=RData.Strings[X];
          if (Fields[X].DataType=ftDate) or
             (Fields[X].DataType=ftTime) or
             (Fields[X].DataType=ftDateTime) Then
          Begin
           P:=Pos(' ',S);
           if P > 0 Then S:=Copy(S,1,P-1);
          End;
          if (F_DaoTable.Fields.Item[X].Attributes And dbUpdatableField) > 0 Then
              Begin
                if (F_DaoTable.Fields.Item[X].Attributes And dbAutoIncrField) = 0 Then
                    Begin
                      if (F_DaoTable.Fields.Item[X].Attributes And dbSystemField) = 0 Then
                          Begin
                            if (Fields[X].IsBlob) And (S = '(Blob)')  Then
                            Else F_DaoTable.Fields.Item[X].Value:=S;
                          End;
                    End;
              End;
        End;
    OleVariant(F_DaoTable).Update;
  End
 Else
  Begin
    Try
      InternalLast;
    Except
      //************************************************************ EXCEPTION NO CURRENT RECORD
    End;
    Buffer:=GetActiveRecordBuffer;
    RData:=PDaoInfo(Buffer+F_StartMyInfo)^.RecordData;
    F_DaoTable.AddNew;
    For X:=0 to RData.Count-1 do
        Begin
          S:=RData.Strings[X];
          if (Fields[X].DataType=ftDate) or                        
             (Fields[X].DataType=ftTime) or
             (Fields[X].DataType=ftDateTime) Then
          Begin                                                        
           P:=Pos(' ',S);
           if P > 0 Then S:=Copy(S,1,P-1);
          End;
          if (F_DaoTable.Fields.Item[X].Attributes And dbUpdatableField > 0) Then
              Begin
                if (F_DaoTable.Fields.Item[X].Attributes And dbAutoIncrField) = 0 Then
                    Begin
                      if (F_DaoTable.Fields.Item[X].Attributes And dbSystemField) = 0 Then
                          Begin
                            if (Fields[X].IsBlob) And (S = '(Blob)')  Then
                            Else F_DaoTable.Fields.Item[X].Value:=S;
                          End;
                    End;
              End;
        End;
    OleVariant(F_DaoTable).Update;
  End;
End;


Procedure TKADaoTable.InternalDelete;
Begin
  F_DaoTable.Delete;
  IF F_DaoTable.EOF then OleVariant(F_DaoTable).MoveLast Else F_DaoTable.MoveNext;
End;


Function TKADaoTable.IsCursorOpen: Boolean;
begin
        Result:=F_Active;
end;

Function TKADaoTable.GetCanModify: Boolean;
begin
        Result:= NOT F_ReadOnly;
end;

Function TKADaoTable.GetRecordSize: Word;
begin
        Result:=F_BufferSize;
end;

Function TKADaoTable.AllocRecordBuffer: PChar;
begin
        GetMem(Result,F_BufferSize);
        FillChar(Result^,F_BufferSize,0);
        PDaoInfo(Result+F_StartMyInfo)^.RecordData:=TStringList.Create;
        PDaoInfo(Result+F_StartMyInfo)^.DaoBookmark.pvData:=Nil;
end;

Procedure TKADaoTable.FreeRecordBuffer(var Buffer: PChar);
begin
        PDaoInfo(Buffer+F_StartMyInfo)^.RecordData.Free;
        FreeMem(Buffer,F_BufferSize);
        Buffer:=Nil;
end;


Function TKADaoTable.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
 Acceptable : Boolean;
 X          : Integer;
 RD         : Variant;
begin
        Result:=grOK;
        Acceptable:=False;
        repeat
                begin
                        case GetMode of
                        gmCurrent:
                                begin
                                   if F_DaoTable.EOF then Result:=grBOF //********************* NEW
                                      Else
                                   if F_DaoTable.BOF then Result:=grEOF; //********************* NEW
                                end;
                        gmNext:
                                begin
                                   if NOT F_DaoTable.EOF Then //********************* NEW
                                        begin
                                          if F_DaoTable.BOF Then F_DaoTable.MoveFirst Else F_DaoTable.MoveNext;
                                          Inc(F_RecNo);
                                          if F_DaoTable.EOF then Result:=grEOF; //********************* NEW
                                        end
                                    else
                                         Begin
                                           Result:=grEOF;
                                         End;
                                end;
                        gmPrior:
                                begin
                                  if F_RecNo > 0 then
                                        begin
                                          if F_DaoTable.EOF Then OleVariant(F_DaoTable).MoveLast Else F_DaoTable.MovePrevious;
                                          Dec(F_RecNo);
                                          if F_DaoTable.BOF then Result:=grBOF;
                                        end
                                   else
                                        begin
                                         Result:=grBOF;
                                        end;
                                end;
                        end;
                        if Result=grOk then
                        begin
                                with PDaoInfo(Buffer+F_StartMyInfo)^ do
                                begin
                                        DaoBookmark:=F_DaoTable.Bookmark^;
                                        GetMem(DaoBookmark.pvData,SizeOF(Integer));
                                        Move(F_DaoTable.Bookmark^.pvData^,DaoBookmark.pvData^,SizeOF(Integer));
                                        RecordNo     := F_RecNo;
                                        BookmarkFlag := bfCurrent;
                                        RecordData.Clear;
                                        For X:=0 To F_DaoTable.Fields.Count-1 do
                                            Begin
                                             if NOT Fields[X].IsBlob Then
                                                RD:=F_DaoTable.Fields.Item[X].Value
                                             Else
                                                RD:='(Blob)';
                                             if VarType(RD) = varNull then RD := '';
                                             RecordData.Add(RD);
                                            End;
                                End;
                                //fill calc fields part of buffer
                                ClearDBCalcFields(Buffer);
                                GetCalcFields(Buffer);
                                Acceptable:=FilterRecord(Buffer);
                                if (GetMode=gmCurrent) and not Acceptable then Result:=grError;
                        end
                end;
        until (Result<>grOk) or Acceptable;
end;

Function TKADaoTable.FilterRecord(Buffer: PChar): Boolean;
var
        SaveState: TDatasetState;
begin
        Result:=True;
        if (not Filtered) or (not Assigned(OnFilterRecord)) then Exit;
        SaveState:=SetTempState(dsFilter);
        F_FilterBuffer:=Buffer;
        if Assigned(OnFilterRecord) then OnFilterRecord(self,Result);
        RestoreState(SaveState);
end;


Function TKADaoTable.GetRecordCount: Integer;
Begin
 Result:=0;
 Try
   F_DaoTable.MoveFirst;
   OleVariant(F_DaoTable).MoveLast;
   Result:=F_DaoTable.RecordCount;
   F_DaoTable.MoveFirst;
   OleVariant(F_DaoTable).Move(F_RecNo);
 Except
 End;
End;



//*******************************************
//*               Blob Stream Handling
//*******************************************


constructor TKBlobStream.Create(Field: TBlobField; Mode: TBlobStreamMode);
Var
   RD     : OleVariant;
begin
     F_BlobData := '';
     F_Mode     := Mode;
     F_Field    := Field;
     F_DataSet  := F_Field.DataSet as TKADaoTable;
     F_Buffer   := F_DataSet.GetActiveRecordBuffer;
     if not (F_Buffer = nil) then F_DataSet.InternalSetToRecord(F_Buffer);
     F_Opened := True;
     if not F_Field.Modified then
     begin
          if (F_DataSet.State = dsInsert)
          then
            Begin                                                  
              F_BlobData := '';
             End
          Else
             Begin
               if Not (F_Buffer = nil) Then
                  Begin
                    Try
                      RD:=F_DataSet.F_DaoTable.Fields.Item[F_Field.FieldNo-1].Value;
                      if VarType(RD) = varNull then RD := '';
                      F_BlobData:=RD;
                     Except
                      F_BlobData:='';
                     End;
                  End
               Else
                  Begin
                    F_BlobData:='';
                  End;
             End;
          F_BlobSize:=Length(F_BlobData);
     end;
     if Mode = bmWrite then Truncate;
end;


destructor TKBlobStream.Destroy;
begin
   if F_Modified then
     try
          F_DataSet.DataEvent(deFieldChange, Longint(F_Field));
          F_BlobData:='';
          F_Buffer:=Nil;
     except
          Application.HandleException(Self);
     end;
end;


function TKBlobStream.Read(var Buffer; Count: Longint): Longint;
begin
     Result := 0;
     if F_Opened then
     begin
          if Count > Size - F_Position then
             Result := Size - F_Position else
             Result := Count;
          if Result > 0 then
          begin
               Move(PChar(F_BlobData)[F_Position], Buffer, Result);
               Inc(F_Position, Result);
          end;
     end;
end;


function TKBlobStream.Write(const Buffer; Count: Longint): Longint;
var
   pTemp  : Pointer;
   RData  : TStringList;
begin
     Result := 0;
     if F_Opened then
     begin
          try
          begin
               SetLength(F_BlobData,Count+1);
               pTemp:=PChar(F_BlobData);
               CopyMemory(pTemp, @Buffer, Count);
               F_BlobData  := Copy(F_BlobData,1,Count);
               F_BlobSize  := Length(F_BlobData);
               RData:=PDaoInfo(F_Buffer+F_DataSet.F_StartMyInfo)^.RecordData;
               RData.Strings[F_Field.FieldNo-1]:=F_BlobData;
               F_Modified := True;
          end;
          finally
          end;
          Inc(F_Position, Count);
          Result := Count;
          F_Modified := True;
     end;
end;


function TKBlobStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
     case Origin of
          0: F_Position := Offset;
          1: Inc(F_Position, Offset);
          2: F_Position := F_BlobSize + Offset;
     end;
     Result := F_Position;
end;


procedure TKBlobStream.Truncate;
begin
     if F_Opened then F_Modified := True;
end;

//***********************************************************************************
Function TKADaoTable.IntegerToBuffer(Buffer: Pointer; S: String): Boolean;
begin
     Result:=False;
     if Buffer=Nil Then Exit;
     Result := (S <> '');
     if S = '' then S := '0';
     Integer(Buffer^) := StrToInt(S);
end;


Function TKADaoTable.FloatToBuffer(Buffer: Pointer; S: String): Boolean;
begin
     Result:=False;
     if Buffer=Nil Then Exit;
     Result := (S <> '');
     if S = '' then S := '0';
     Double(Buffer^) := StrToFloat(S);
end;


Function TKADaoTable.DateToBuffer(Buffer: Pointer; S: String): Boolean;
var
   Ttmp : TTimeStamp;
   dTmp : ^TDateTimeRec;
   P    : Integer;
   SS   : String;
begin
     Result:=False;
     if Buffer=Nil Then Exit;
     Result := (S <> '');
     P:=Pos(' ',S);
     if P > 0 Then SS:=Copy(S,1,P-1);
     System.Delete(S,1,P);
     P:=Pos(' ',S);
     if P > 0 Then
        Begin
         System.Delete(S,1,P);
         S:=SS+' '+S;
        End
     Else
        Begin
         S:=SS;
        End;
     Ttmp.Date := 0;
     Ttmp.Time := 0;
     if S <> '' then Ttmp := DateTimeToTimeStamp(StrToDateTime(S));
     Dtmp := Buffer;
     Dtmp.Date := Ttmp.Date;
end;


Function TKADaoTable.TimeToBuffer(Buffer: Pointer; S: String): Boolean;
var
   Ttmp : TTimeStamp;
   Dtmp : ^TDateTimeRec;
begin
     Result:=False;
     if Buffer=Nil Then Exit;
     Result := (S <> '');
     Ttmp.Date := 0;
     Ttmp.Time := 0;
     if S <> '' then
     begin
          if Length(S) > 10
          then
              Ttmp := DateTimeToTimeStamp(StrToDateTime(S))
          else
              Ttmp := DateTimeToTimeStamp(StrToDate(S));
     end;
     Dtmp := Buffer;
     Dtmp.Time := Ttmp.Time;
end;


Function TKADaoTable.DateTimeToBuffer(Buffer: Pointer; S: String): Boolean;
var
   Ttmp : TTimeStamp;
   Dtmp : ^TDateTimeRec;
begin
     Result:=False;
     if Buffer=Nil Then Exit;
     Result := (S <> '');
     Ttmp.Date := 0;
     Ttmp.Time := 0;
     if S <> '' then Ttmp := DateTimeToTimeStamp(StrToDate(S));
     Dtmp := Buffer;
     Dtmp^.DateTime := TimeStampToMSecs(Ttmp);
end;


Function TKADaoTable.BooleanToBuffer(Buffer: Pointer; S: String): Boolean;
begin
     Result:=False;
     if Buffer=Nil Then Exit;
     Result := (S <> '');
     if S = ''    Then S := '0';
     if S = '-1'  Then S := '1';
     if AnsiLowerCase(S)='false' Then S := '0';
     if AnsiLowerCase(S)='true'  Then S := '1';
     WordBool(Buffer^) := WordBool(StrToInt(S));
end;


Function TKADaoTable.BufferToDate(Buffer: Pointer): String;
var
   Ttmp : TTimeStamp;
   Dtmp : ^TDateTimeRec;
begin
     Dtmp := Buffer;
     Ttmp.Date := Dtmp^.Date;
     Result := DateToStr(TimeStampToDateTime(Ttmp));
end;


Function TKADaoTable.BufferToTime(Buffer: Pointer): String;
var
   TTmp : TTimeStamp;
   Dtmp : ^TDateTimeRec;
begin
     Dtmp := Buffer;
     Ttmp.Time := Dtmp^.Time;
     Result := FormatDateTime('hh:nn:ss', TimeStampToDateTime(Ttmp));
end;


Function TKADaoTable.BufferToDateTime(Buffer: Pointer): String;
var
   TTmp : TTimeStamp;
   Dtmp : ^TDateTimeRec;
begin
     Dtmp   := Buffer;
     Ttmp   := MsecsToTimeStamp(Dtmp.DateTime);
     Result := DateToStr(TimeStampToDateTime(TTmp));
end;



//*********************************************************************************** TableName
Procedure TTableNameEditor.GetValues( Proc: TGetStrProc);
Var
  DTable : TKADaoTable;
  DBase : TKADaodatabase;
  X     : Integer;
Begin
  if GetComponent(0) is TKADaoTable then
  Begin
    DTable := TKADaoTable(GetComponent(0));
    if Assigned(DTable.F_Database) Then
       Begin
        DBase:=DTable.F_Database;
        Try
          For X := 0 to DBase.TableNames.Count-1 do Proc(DBase.TableNames.Strings[X]);
        Finally
        End;
       End;
     Modified;
  End;
End;

Function TTableNameEditor.GetAttributes: TPropertyAttributes;
Begin
  Result:= Inherited GetAttributes + [paValueList, paSortList];
End;
//*********************************************************************************** QueryName

Procedure TQueryNameEditor.GetValues( Proc: TGetStrProc);
Var
  DTable : TKADaoTable;
  DBase : TKADaodatabase;
  X     : Integer;
Begin
  if GetComponent(0) is TKADaoTable then
  Begin
    DTable := TKADaoTable(GetComponent(0));
    if Assigned(DTable.F_Database) Then
       Begin
        DBase:=DTable.F_Database;
        Try
          For X := 0 to DBase.QueryDefNames.Count-1 do Proc(DBase.QueryDefNames.Strings[X]);
        Finally
        End;
       End;
     Modified;
  End;
End;

Function TQueryNameEditor.GetAttributes: TPropertyAttributes;
Begin
  Result:= Inherited GetAttributes + [paValueList, paSortList];
End;
//*********************************************************************************** IndexName
Procedure TIndexNameEditor.GetValues( Proc: TGetStrProc);
Var
  DTable : TKADaoTable;
  DBase  : TKADaodatabase;
  X      : Integer;
  Count  : Integer;
Begin
  if GetComponent(0) is TKADaoTable then
  Begin
    DTable := TKADaoTable(GetComponent(0));
    if Assigned(DTable.F_Database) Then
    if DTable.TableName <> '' Then
       Begin
        DBase:=DTable.F_Database;
        Try
          Count :=DBase.CoreDatabase.TableDefs.Item[DTable.TableName].Indexes.Count;
          For X := 0 to  Count-1 do
              Begin
                Proc(DBase.CoreDatabase.TableDefs.Item[DTable.TableName].Indexes.Item[X].Name);
              End;
        Finally
        End;
       End;
     Modified;
  End;
End;


Function TIndexNameEditor.GetAttributes: TPropertyAttributes;
Begin
  Result:= Inherited GetAttributes + [paValueList, paSortList];
End;

//***************************************************************************************************************************

Function TTableTypeEditor.GetAttributes: TPropertyAttributes;
Begin
  Result := Inherited GetAttributes + [paValueList, paSortList];
End;

Function  TTableTypeEditor.GetValue: string;
Begin
 if GetComponent(0) is TKADaoTable then
    Begin
      Result:='StandardTable';
      if TKADaoTable(GetComponent(0)).F_TableType=dbOpenTable       then Result:='StandardTable';
      if TKADaoTable(GetComponent(0)).F_TableType=dbOpenDynaset     then Result:='DynasetTable';
      if TKADaoTable(GetComponent(0)).F_TableType=dbOpenDynamic     then Result:='DynamicTable';
      if TKADaoTable(GetComponent(0)).F_TableType=dbOpenSnapshot    then Result:='SnapshotTable';
      if TKADaoTable(GetComponent(0)).F_TableType=dbOpenForwardOnly then Result:='ForwardOnlyTable';
    End;
End;

Procedure TTableTypeEditor.GetValues( Proc: TGetStrProc);
Begin
  if GetComponent(0) is TKADaoTable then
     Begin
       Proc('StandardTable');
       Proc('DynasetTable');
       Proc('DynamicTable');
       Proc('SnapshotTable');
       Proc('ForwardOnlyTable');
       Modified;
     End;
End;


Procedure TTableTypeEditor.SetValue(const Value: string);
Var
 Dat : Integer;
Begin
  if GetComponent(0) is TKADaoTable then
       Begin
       Dat:=dbOpenTable;
       if Value='StandardTable'    Then Dat:=dbOpenTable;
       if Value='DynasetTable'     Then Dat:=dbOpenDynaset;
       if Value='DynamicTable'     Then Dat:=dbOpenDynamic;
       if Value='SnapshotTable'    Then Dat:=dbOpenSnapshot;
       if Value='ForwardOnlyTable' Then Dat:=dbOpenForwardOnly;
       Inherited SetValue(IntToStr(Dat));
       Modified;
     End;
End;

//***************************************************************************************************************************

Function TLockTypeEditor.GetAttributes: TPropertyAttributes;
Begin
  Result := Inherited GetAttributes + [paValueList, paSortList];
End;

Function  TLockTypeEditor.GetValue: string;
Begin
 if GetComponent(0) is TKADaoTable then
    Begin
      Result:='dbPessimistic';
      {$IFDEF DAO35}
      if TKADaoTable(GetComponent(0)).F_LockType=DAO35Api.dbPessimistic       then Result:='dbPessimistic';
      if TKADaoTable(GetComponent(0)).F_LockType=DAO35Api.dbOptimistic        then Result:='dbOptimistic';
      if TKADaoTable(GetComponent(0)).F_LockType=DAO35Api.dbOptimisticValue   then Result:='dbOptimisticValue';
      if TKADaoTable(GetComponent(0)).F_LockType=DAO35Api.dbReadOnly          then Result:='dbReadOnly';
      if TKADaoTable(GetComponent(0)).F_LockType=DAO35Api.dbOptimisticBatch   then Result:='dbOptimisticBatch';
      {$ENDIF}
      {$IFDEF DAO36}
      if TKADaoTable(GetComponent(0)).F_LockType=DAO36Api.dbPessimistic       then Result:='dbPessimistic';
      if TKADaoTable(GetComponent(0)).F_LockType=DAO36Api.dbOptimistic        then Result:='dbOptimistic';
      if TKADaoTable(GetComponent(0)).F_LockType=DAO36Api.dbOptimisticValue   then Result:='dbOptimisticValue';
      if TKADaoTable(GetComponent(0)).F_LockType=DAO36Api.dbReadOnly          then Result:='dbReadOnly';
      if TKADaoTable(GetComponent(0)).F_LockType=DAO36Api.dbOptimisticBatch   then Result:='dbOptimisticBatch';
      {$ENDIF}
    End;
End;

Procedure TLockTypeEditor.GetValues( Proc: TGetStrProc);
Begin
  if GetComponent(0) is TKADaoTable then
     Begin
       Proc('dbPessimistic');
       Proc('dbOptimistic');
       Proc('dbOptimisticValue');
       Proc('dbReadOnly');
       Proc('dbOptimisticBatch');
       Modified;
     End;
End;


Procedure TLockTypeEditor.SetValue(const Value: string);
Var
 Dat : Integer;
Begin
  if GetComponent(0) is TKADaoTable then
       Begin
       {$IFDEF DAO35}
       Dat:=DAO35Api.dbPessimistic;
       if Value='dbPessimistic'     Then Dat:=DAO35Api.dbPessimistic;
       if Value='dbOptimistic'      Then Dat:=DAO35Api.dbOptimistic;
       if Value='dbOptimisticValue' Then Dat:=DAO35Api.dbOptimisticValue;
       if Value='dbReadOnly'        Then Dat:=DAO35Api.dbReadOnly;
       if Value='dbOptimisticBatch' Then Dat:=DAO35Api.dbOptimisticBatch;
       {$ENDIF}
       {$IFDEF DAO36}
       Dat:=DAO36Api.dbPessimistic;
       if Value='dbPessimistic'     Then Dat:=DAO36Api.dbPessimistic;
       if Value='dbOptimistic'      Then Dat:=DAO36Api.dbOptimistic;
       if Value='dbOptimisticValue' Then Dat:=DAO36Api.dbOptimisticValue;
       if Value='dbReadOnly'        Then Dat:=DAO36Api.dbReadOnly;
       if Value='dbOptimisticBatch' Then Dat:=DAO36Api.dbOptimisticBatch;
       {$ENDIF}
       Inherited SetValue(IntToStr(Dat));
       Modified;
     End;
End;
//***************************************************************************************************************************
Procedure Register;
begin
    RegisterComponents('KA Dao', [TKADaoTable]);
    RegisterPropertyEditor(TypeInfo(String),TKADaoTable,'TableName',TTableNameEditor);
    RegisterPropertyEditor(TypeInfo(String),TKADaoTable,'QueryDefName',TQueryNameEditor);
    RegisterPropertyEditor(TypeInfo(String),TKADaoTable,'IndexName',TIndexNameEditor);
    RegisterPropertyEditor(TypeInfo(Integer),TKADaoTable,'TableType',TTableTypeEditor);
    RegisterPropertyEditor(TypeInfo(Integer),TKADaoTable,'LockType',TLockTypeEditor);
end;


end.

